/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         ecall_control.c                                              */
/* Version:      8.6.0 (Rel8) / 9.4.0 (Rel9)                                  */
/* Date:         2011-02-08                                                   */
/* Description:  control message functions                                    */
/*----------------------------------------------------------------------------*/

#include "modemx.h"


/*============================================================================*/
/* UTILITY FUNCTION: CtrlPortName                                             */
/*----------------------------------------------------------------------------*/
/* Description: converter for log output                                      */
/*                                                                            */
/* In:     PortOwner   owner  -> modem identification                         */
/* Return: const char*        <- port name as character string                */
/*----------------------------------------------------------------------------*/

const char* CtrlPortName(PortOwner owner)
{
  static const char* strIvs   = "IVS";
  static const char* strPsap  = "PSAP";
  static const char* strUndef = "VOID";

  switch (owner) {
    case PortIvs:
      return strIvs;
      break;
    case PortPsap:
      return strPsap;
      break;
    default:
      return strUndef;
      break;
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: CtrlSync                                                 */
/*----------------------------------------------------------------------------*/
/* Description: control message sync function                                 */
/*                                                                            */
/* InOut:  CtrlRxData*  control  <-> control struct                           */
/* In:     const Int16* pcm       -> input frame of 16bit PCM samples         */
/*----------------------------------------------------------------------------*/

void CtrlSync(CtrlRxData *control, const Int16 *pcm)
{
  CtrlPort *port = &control->port;
  SyncState *sync = control->sync;

  /* initialize port data */
  port->u.rx.dlData   = DlMsgNoop;
  port->u.rx.dlMetric = 0;

  if (sync->events < control->dlSyncLock) {
    sync->delay = 0;

    /* keep control receiver inactive while sync is running */
    if (sync->resync == False) {
      port->state = CtrlRxIdle;
    }
    /* run synchronization */
    Sync(sync, pcm, CtrlPortName(port->owner), port->invert);

    if (sync->flag == True && sync->delay <= 0) {
      if (
        sync->index >= NRF_DLCHUNK ||
        sync->amplitude[2] < sync->amplitude[0])
      {
        /* change port state to sync received */
        port->state = CtrlRxSync;

        switch (port->owner) {
          case PortIvs:
            IvsCatchEvent(IVSEVENT_CONTROLSYNC);
            break;
          case PortPsap:
            PsapCatchEvent(PSAPEVENT_CONTROLSYNC);
            break;
          default:
            LogInfo("[%-4s] control callback: port owner undefined.",
              CtrlPortName(PortUndef));
            break;
        }
        /* count occurrences of the same sync position */
        if (sync->events > 0 &&
          sync->delayMem == sync->delay && (sync->index)%NRF_DLCHUNK == 0)
        {
          sync->events++;
          sync->index = 0;
        }
        else {
          sync->events   = 1;
          sync->index    = 0;
          sync->delayMem = sync->delay;
        }
        if (sync->events >= control->dlSyncLock) {
          /* check for muting gaps passed already (avoids unnecessary delays) */
          while (sync->delay <= -PCM_LENGTH) {
            sync->delay += PCM_LENGTH;
            sync->index++;
          }
          sync->delayMem   = sync->delay;
          control->dlIndex = sync->index;

          if (sync->resync == False) {
            control->dlRead = Plus;
            control->dlSyncLock = LOCK_RESYNC;

            /* lock sync for control message detection */
            port->state = CtrlRxLock;

            switch (port->owner) {
              case PortIvs:
                IvsCatchEvent(IVSEVENT_CONTROLLOCK);
                break;
              case PortPsap:
                PsapCatchEvent(PSAPEVENT_CONTROLLOCK);
                break;
              default:
                LogInfo("[%-4s] control callback: port owner undefined.",
                  CtrlPortName(PortUndef));
                break;
            }
            LogInfo("[%-4s] sync locked, starting control message detection",
              CtrlPortName(port->owner));

            /* apply sync inversion flag */
            if ((port->invert = sync->invert) == True) {
              LogInfo("[%-4s] codec inversion detected!",
                CtrlPortName(port->owner));
            }
          }
          else if (sync->invert == port->invert) {
            /* resynchronization event (assume inversion has not changed) */
            control->dlRead = Plus;
          }
          else {
            control->dlRead = Minus;
          }
        } else {
          sync->amplitude[2] = sync->amplitude[0];
          sync->amplitude[0] = 0;
          sync->npeaks = 0;
        }
      } else {
        sync->amplitude[0] = 0;
        sync->npeaks = 0;
      }
    } else if (sync->resync == True && sync->index > NRF_RESYNC) {
      port->u.rx.dlData   = DlMsgReset;
      port->u.rx.dlMetric = 0;
    }
  }
  else {
    /* sync is locked, check whether locked sync is still valid */
    port->state = CtrlRxSync;
    SyncCheck(sync, pcm, CtrlPortName(port->owner));

    if (sync->flag == True) {
      control->dlIndex = sync->index = (sync->index)%NRF_DLCHUNK;

      while (sync->delayMem <= -PCM_LENGTH) {
        sync->delayMem += PCM_LENGTH;
        sync->index++;
        control->dlIndex++;
      }
      while (sync->delayMem > 0) {
        sync->delayMem -= PCM_LENGTH;
        sync->index--;
        control->dlIndex--;
      }

      if (sync->check == Plus) {
        /* lower-layer control message */
        control->dlRead = Plus;
        sync->checkCnt = 0;
      }
      else if (sync->check == Minus) {
        /* higher-layer control message */
        control->dlRead = Minus;
        sync->checkCnt = 0;
      }
      else {
        /* following message is ignored */
        control->dlRead = Zero;
        sync->checkCnt++;
        LogInfo("[%-4s] message ignored (bad sync check)",
          CtrlPortName(port->owner));
      }
    }
    /* sync completely lost, reset synchronization */
    if (sync->checkCnt >= SYNC_BADCHECK) {
      SyncReset(sync, NULL, NULL);
      sync->resync = True;
    }
  }
}

/*============================================================================*/
/* CONTROL FUNCTION: CtrlTxReset                                              */
/*----------------------------------------------------------------------------*/
/* Description: reset function control transmitter                            */
/*                                                                            */
/* InOut:  CtrlTxData* control  <-> control struct                            */
/* In:     PortOwner   owner     -> modem identification                      */
/*----------------------------------------------------------------------------*/

void CtrlTxReset(CtrlTxData *control, PortOwner owner)
{
  control->port.state  = CtrlTxIdle;
  control->port.invert = False;
  control->port.owner  = owner;

  /* initial port data */
  control->port.u.tx.dlData  = DlMsgNoop;
  control->port.u.tx.dlIndex = -1;
}

/*============================================================================*/
/* CONTROL FUNCTION: CtrlTxProcess                                            */
/*----------------------------------------------------------------------------*/
/* Description: process function control transmitter                          */
/*                                                                            */
/* InOut:  CtrlTxData* control  <-> control struct                            */
/*         Int16* pcm           <-> frame of 16bit PCM samples                */
/*----------------------------------------------------------------------------*/

void CtrlTxProcess(CtrlTxData *control, Int16 *pcm)
{
  CtrlPort *port = &control->port;

  /* PSAP transmitter, initialize PCM data to zero */
  memset(pcm, 0, sizeof(Int16)*PCM_LENGTH);

  /* safety check on index range */
  if (port->u.tx.dlIndex >= 0 && port->u.tx.dlIndex < NRF_DLCHUNK) {
    if (port->u.tx.dlData != DlMsgNoop) {
      CtrlTxMod(pcm, port->u.tx.dlData, port->u.tx.dlIndex);
    }
  } else if (port->u.tx.dlIndex >= NRF_DLCHUNK) {
    Abort("message index too large (value: %d), exit.", port->u.tx.dlIndex);
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: CtrlTxMod                                                */
/*----------------------------------------------------------------------------*/
/* Description: control message transmitter using prestored sequences         */
/*                                                                            */
/* In:     Int16  symbol  -> lower-layer or higher-layer message symbol       */
/*         Int16  index   -> position within message frame                    */
/* Out:    Int16* pcm     <- output data                                      */
/*----------------------------------------------------------------------------*/

void CtrlTxMod(Int16 *pcm, Int16 symbol, Int16 index)
{
  Int16 i;
  Int16 lpcm = sizeof(Int16)*PCM_LENGTH;

  if (!(symbol >= 0 && symbol < 4) &&
    !(symbol >= SET_LLMSG && symbol < SET_LLMSG + SET_HLMSG))
  {
    Abort("message symbol out of range, exit.");
  }
  if (index < NRF_WAKEUP) {
    /* wakeup tone */
    for (i = 0; i < PCM_LENGTH; i++)
      pcm[i] = wakeupSin500[i%16];
  }
  else if (index < NRF_SYNC) {
    /* sync frame */
    memcpy(pcm, &syncFrame[(index-NRF_WAKEUP)*PCM_LENGTH], lpcm);

    /* invert sync preamble for higher-layer messages */
    if (symbol >= SET_LLMSG) {
      for (i = 0; i < PCM_LENGTH; i++)
        pcm[i] = -pcm[i];
    }
    /* add signal offset */
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] += (pcm[i] == 0) ? 12000 : 5000;
    }

    /* wakeup tone extends into the first frame of the preamble */
    if (NRF_WAKEUP > 0 && index == NRF_WAKEUP) {
      for (i = 0; i < 32; i++)
        pcm[i] = wakeupSin500[i%16];
    }
  }
  else {
    i = (symbol < SET_LLMSG)
      ? index - NRF_DLMUTE1LL - NRF_SYNC
      : index - NRF_DLMUTE1HL - NRF_SYNC;

    if (symbol < SET_LLMSG && i >= 0 && i < NRF_DLDATA) {
      /* data message lower-layer symbol */
      memcpy(pcm, &dlPcmData[symbol][i*PCM_LENGTH], lpcm);
    }
    else if (symbol >= SET_LLMSG && i >= 0 && i < NRF_DLDATA) {
      /* 1st part of data message higher-layer symbol */
      memcpy(pcm, &dlPcmData[(symbol-SET_LLMSG)%4][i*PCM_LENGTH], lpcm);
    }
    else if (symbol >= SET_LLMSG && i >= NRF_DLDATA && i < 2*NRF_DLDATA) {
      /* 2nd part of data message higher-layer symbol */
      i -= NRF_DLDATA;
      memcpy(pcm, &dlPcmData[((symbol-SET_LLMSG)/4)%4][i*PCM_LENGTH], lpcm);
    }
    else {
      /* muting frame */
      memset(pcm, 0, lpcm);
    }
  }
}

/*============================================================================*/
/* CONTROL FUNCTION: CtrlRxReset                                              */
/*----------------------------------------------------------------------------*/
/* Description: reset function control receiver                               */
/*                                                                            */
/* InOut:  CtrlRxData* control   <-> control struct                           */
/* In:     PortOwner   owner      -> modem identification                     */
/*         SyncState*  sync       -> pointer to sync struct                   */
/*         Int16*      buffer     -> pointer to control receiver buffer       */
/*         Int16       syncLock   -> number of sync events required           */
/*----------------------------------------------------------------------------*/

void CtrlRxReset(CtrlRxData *control, PortOwner owner,
                 SyncState *sync, Int16 *buffer, Int16 syncLock)
{
  control->port.state  = CtrlRxIdle;
  control->port.invert = False;
  control->port.owner  = owner;

  /* initial port data */
  control->port.u.rx.dlData   = DlMsgNoop;
  control->port.u.rx.dlMetric = 0;

  /* set control pointers */
  control->sync   = sync;
  control->buffer = buffer;

  /* set internal variables */
  control->dlRead     = Zero;
  control->dlIndex    = 0;
  control->dlSyncLock = syncLock;

  /* clear memory */
  memset(control->buffer, 0, sizeof(Int16)*(NRF_MEMCTRL*PCM_LENGTH));
}

/*============================================================================*/
/* CONTROL FUNCTION: CtrlRxProcess                                            */
/*----------------------------------------------------------------------------*/
/* Description: process function control receiver                             */
/*                                                                            */
/* InOut:  CtrlRxData*  control  <-> control struct                           */
/* In:     const Int16* pcm       -> input frame of 16bit PCM samples         */
/*----------------------------------------------------------------------------*/

void CtrlRxProcess(CtrlRxData *control, const Int16 *pcm)
{
  Int16 i, metric;
  Bool syncRun = False;

  CtrlPort *port = &control->port;
  SyncState *sync = control->sync;

  /* increase sync index */
  sync->index++;

  if (control->dlRead == Zero) {
    /* run synchronization (sync function to set port data) */
    CtrlSync(control, pcm);
    syncRun = True;
  }

  if (control->dlRead == Plus &&
    (control->dlIndex)%NRF_DLCHUNK == NRF_DLMUTE1LL + NRF_DLDATA - 1)
  {
    /* enough lower-layer data received, detect data now */
    i = (NRF_MEMCTRL - NRF_DLDATA)*PCM_LENGTH;
    port->u.rx.dlData =
      CtrlRxDemod(&control->buffer[i + sync->delayMem], &metric);
    port->u.rx.dlMetric = metric;

    /* prepare for next sync */
    sync->index      = (sync->index)%NRF_DLCHUNK;
    sync->check      = Zero;
    control->dlRead  = Zero;
    control->dlIndex = 0;
  }
  else if (control->dlRead == Minus &&
    (control->dlIndex)%NRF_DLCHUNK == NRF_DLMUTE1HL + 2*NRF_DLDATA - 1)
  {
    /* enough higher-layer data received, detect first data portion */
    i = (NRF_MEMCTRL - 2*NRF_DLDATA)*PCM_LENGTH;
    port->u.rx.dlData =
      CtrlRxDemod(&control->buffer[i + sync->delayMem], &metric);
    port->u.rx.dlMetric = metric;

    /* detect second data portion and combine */
    i = (NRF_MEMCTRL - NRF_DLDATA)*PCM_LENGTH;
    port->u.rx.dlData +=
      4*CtrlRxDemod(&control->buffer[i + sync->delayMem], &metric);
    port->u.rx.dlData += SET_LLMSG;
    port->u.rx.dlMetric = MIN(port->u.rx.dlMetric, metric);

    /* prepare for next sync */
    sync->index      = (sync->index)%NRF_DLCHUNK;
    sync->check      = Zero;
    control->dlRead  = Zero;
    control->dlIndex = 0;
  }
  else if (syncRun == False) {
    /* sync detected, wait for enough data to process */
    control->dlIndex++;
    port->u.rx.dlData   = DlMsgNoop;
    port->u.rx.dlMetric = 0;
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: CtrlRxDemod                                              */
/*----------------------------------------------------------------------------*/
/* Description: control message receiver                                      */
/*                                                                            */
/* In:     const Int16* pcm     -> input PCM buffer                           */
/* Out:    Int16*       metric  <- reliability factor (-1: skip)              */
/* Return: Int16                <- demodulated message                        */
/*----------------------------------------------------------------------------*/

Int16 CtrlRxDemod(const Int16 *pcm, Int16 *metric)
{
  Int16 k, n, msg;
  Int32 xCorr[4];
  Int32 maxCorr, runnerup;

  /* correlation receiver */
  for (k = 0; k < 4; k++) {
    xCorr[k] = 0;
    for (n = 0; n < NRF_DLDATA*PCM_LENGTH; n++)
      xCorr[k] += (pcm[n]*dlPcmDataMatch[k][n]) >> 12;
  }

  /* find largest correlation value */
  maxCorr = xCorr[0];
  msg = 0;
  for (k = 1; k < 4; k++) {
    if (xCorr[k] > maxCorr) {
      maxCorr = xCorr[k];
      msg = k;
    }
  }
  /* find second largest correlation value */
  runnerup = xCorr[(msg+1)%4];
  for (k = 0; k < 4; k++) {
    if (xCorr[k] > runnerup && xCorr[k] < maxCorr)
      runnerup = xCorr[k];
  }
  /* test on reliability by absolute and relative value */
  if (maxCorr < IVS_THRESHOLD) {
    *metric = -1;
  }
  else if ((*metric = (Int16)((maxCorr-runnerup)/(maxCorr/100))) <= 33) {
    *metric = -1;
  }
  return msg;
}
