/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         modem_ivs.c                                                  */
/* Version:      8.6.0 (Rel8) / 9.4.0 (Rel9)                                  */
/* Date:         2011-02-08                                                   */
/* Description:  eCall IVS modem                                              */
/*----------------------------------------------------------------------------*/

#include "modemx.h"


static IvsData ivs;

/* IVS external reference */
IvsData* IvsGetData(void) { return &ivs; }

/*============================================================================*/
/* IVS IMPLEMENTATION: IvsReset, IvsSendStart                                 */
/*----------------------------------------------------------------------------*/
/* In:  const Ord8* msd     -> MSD to be transmitted                          */
/*      int         length  -> MSD length (equal to MSD_MAX_LENGTH)           */
/*----------------------------------------------------------------------------*/

void IvsReset(const Ord8 *msd, int length) {
  IvsRxReset();
  IvsTxReset(msd, length);
}

void IvsSendStart(void) {
  if (ivs.rx.state == IvsIdle) {
    LogInfo("[IVS ] SendStart received from control");
    LogInfo("[----]");

    ivs.rx.state = IvsTrigger;
    IvsCatchEvent(IVSEVENT_SENDINGSTART);
  }
  else {
    /* ignore otherwise */
    LogInfo("[IVS ] SendStart received from control, ignored.");
    LogInfo("[----]");
  }
}

/*----------------------------------------------------------------------------*/

void IvsRxReset(void) {
  CtrlRxData *ctrl = &ivs.rx.ctrl;
  SyncState  *sync = &ivs.rx.sync;

  /* reset state */
  ivs.rx.state = IvsIdle;

  /* reset control receiver */
  CtrlRxReset(ctrl, PortIvs, sync, ivs.rx.memCtrl, LOCK_START_DL);

  /* reset counters for SEND uplink control */
  ivs.rx.dlData   = DlMsgNoop;
  ivs.rx.dlMsgCnt = -1;
  ivs.rx.dlIndex  = -1;

  /* reset sync */
  SyncReset(sync, ivs.rx.memSync, NULL);

  LogInfo("[IVS ] reset receiver");
}

/*----------------------------------------------------------------------------*/

void IvsTxReset(const Ord8 *msd, int length) {
  Int16 k;

  /* reset state */
  ivs.tx.state = IvsIdle;

  /* set control link parameters (SEND) */
  CtrlTxReset(&ivs.tx.ctrl, PortIvs);

  /* reset state counters */
  for (k = 0; k <= SET_LLMSG + SET_HLMSG; k++) {
    ivs.tx.stateCnt[k] = 0;
    ivs.tx.stateIgn[k] = 0;
  }
  ivs.tx.stateCnt[DlMsgSend] = -1;
  ivs.tx.pendingStart = False;

  /* reset uplink counters and message memory */
  ivs.tx.rv       = 0;
  ivs.tx.ulN      = 0;
  ivs.tx.ulIndex  = 0;
  ivs.tx.ulDelay  = 73;
  ivs.tx.dlMsgOld = DlMsgNoop;

  if (msd != NULL) {
    /* initialize modulator and clear cumulative NACK count */
    SetModState(&ivs.tx.mod, ModUndef);
    ivs.tx.overallNack = 0;

    /* convert MSD and encode (all redundancy versions) */
    Byte2Bit(msd, ivs.tx.memCode, MSD_MAX_LENGTH);
    FecEncode(ivs.tx.memCode);

    LogInfo("[IVS ] reset transmitter (MSD provided)");
  } else {
    LogInfo("[IVS ] reset transmitter");
  }
}

/*============================================================================*/
/* IVS IMPLEMENTATION: IvsProcess                                             */
/*----------------------------------------------------------------------------*/
/* InOut:  Int16* pcm  <-> input and output frame of 16bit PCM samples        */
/*----------------------------------------------------------------------------*/

void IvsProcess(Int16 *pcm) {
  Int16 i;
  CtrlPort *prx = &ivs.rx.ctrl.port;
  CtrlPort *ptx = &ivs.tx.ctrl.port;

  /* invert input signal if required */
  if (prx->invert == True) {
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] = (pcm[i] > PCM_MIN) ? -pcm[i] : PCM_MAX;
    }
  }
  IvsRxProcess(pcm);

  if (prx->state == CtrlTxIdle || prx->state == CtrlTxSend) {
    /* control transmit mode */
    ptx->state        = prx->state;
    ptx->u.tx.dlData  = prx->u.tx.dlData;
    ptx->u.tx.dlIndex = prx->u.tx.dlIndex;
  }
  else {
    /* control receive mode */
    ptx->state         = prx->state;
    ptx->u.rx.dlData   = prx->u.rx.dlData;
    ptx->u.rx.dlMetric = prx->u.rx.dlMetric;
  }
  IvsTxProcess(pcm);

  /* invert output signal if required */
  if (ptx->invert == True) {
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] = (pcm[i] > PCM_MIN) ? -pcm[i] : PCM_MAX;
    }
  }
}

/*----------------------------------------------------------------------------*/

void IvsRxProcess(const Int16 *pcm) {
  Int16 i;
  CtrlPort *port = &ivs.rx.ctrl.port;

  /* IVS receiver, copy PCM data to input buffer */
  for (i = 0; i < NRF_MEMCTRL-1; i++) {
    memcpy(&ivs.rx.memCtrl[i*PCM_LENGTH],
      &ivs.rx.memCtrl[(i+1)*PCM_LENGTH], sizeof(Int16)*PCM_LENGTH);
  }
  memcpy(&ivs.rx.memCtrl[(NRF_MEMCTRL-1)*PCM_LENGTH], pcm,
    sizeof(Int16)*PCM_LENGTH);

  /* IVS receiver, run control message receiver */
  CtrlRxProcess(&ivs.rx.ctrl, pcm);

  /* check for uplink SEND trigger */
  switch (ivs.rx.state) {
    case IvsTrigger:
      if (port->state == CtrlRxIdle) {
        port->state = CtrlTxIdle;

        /* set control data */
        ivs.rx.state    = IvsStart;
        ivs.rx.dlData   = DlMsgSend;
        ivs.rx.dlMsgCnt = 0;
        ivs.rx.dlIndex  = 0;
      }
      else {
        /* SEND trigger not allowed during MSD transmission */
        ivs.rx.state = IvsIdle;
        LogInfo("[IVS ] SendStart during MSD transmission, ignored.");
      }
      break;
    case IvsStart:
      if (port->state == CtrlRxSync || ivs.rx.dlMsgCnt >= IVS_NUMSEND) {
        /* stop SEND trigger */
        ivs.rx.state    = IvsIdle;
        ivs.rx.dlData   = DlMsgNoop;
        ivs.rx.dlMsgCnt = -1;
        ivs.rx.dlIndex  = -1;

        if (ivs.rx.dlMsgCnt >= IVS_NUMSEND) {
          IvsCatchEvent(IVSEVENT_IDLEPOSTSTART);
        }
      }
      else {
        /* sending control data */
        port->state        = CtrlTxSend;
        port->u.tx.dlData  = ivs.rx.dlData;
        port->u.tx.dlIndex = ivs.rx.dlIndex;

        if (ivs.rx.dlIndex++ == 0) {
          LogInfo("[IVS ] sending  SEND");
        }
        else if (ivs.rx.dlIndex == NRF_DLCHUNK) {
          ivs.rx.dlIndex = 0;
          ivs.rx.dlMsgCnt++;
        }
      }
      break;
    default:
      /* do nothing but receiving downlink data */
      break;
  }
}

/*----------------------------------------------------------------------------*/

void IvsTxProcess(Int16 *pcm) {
  Int16 lpcm = sizeof(Int16)*PCM_LENGTH;
  CtrlTxData *ctrl = &ivs.tx.ctrl;

  /* IVS transmitter, initialize PCM data to zero */
  memset(pcm, 0, lpcm);

  if (ctrl->port.state == CtrlTxIdle) {
    /* SEND trigger initiated */
  }
  else if (ctrl->port.state == CtrlTxSend) {
    /* transmit control data on uplink */
    CtrlTxProcess(ctrl, pcm);
  }
  else {
    /* sync loss may trigger a reset */
    if (ctrl->port.u.rx.dlData == DlMsgReset) {
      LogInfo("[IVS ] full reset (no sync detected)");
      IvsReset(NULL, 0);
      IvsCatchEvent(IVSEVENT_IDLEPOSTRESET);
    }
    else if (ctrl->port.u.rx.dlData != DlMsgNoop) {
      /* run state machine based on new downlink symbol */
      IvsTxState(ctrl->port.u.rx.dlData, ctrl->port.u.rx.dlMetric,
        (ctrl->port.state == CtrlRxLock) ? True : False);
    }

    /* run transmitter based on IVS state */
    if (ivs.tx.state == IvsIdle ||
      ivs.tx.state == IvsAck || ivs.tx.state == IvsHlack)
    {
      /* do nothing */
    }
    else if (ivs.tx.state == IvsSendMsd && ivs.tx.mod.type != ModUndef) {
      if (ivs.tx.ulIndex < ivs.tx.ulN) {
        /* run transmitter */
        IvsTransmitter(
          &ivs.tx.mod, ivs.tx.memCode, pcm, ivs.tx.rv, ivs.tx.ulIndex);
        ivs.tx.ulIndex++;
      }
      else {
        /* increase redundancy version counter */
        ivs.tx.rv++;
        ivs.tx.ulIndex = 0;

        /* no sync frame is transmitted next */
        ivs.tx.ulN = ivs.tx.mod.nfmuteall + ivs.tx.mod.nfdata;

        /* new redundancy version, run transmitter */
        IvsTransmitter(
          &ivs.tx.mod, ivs.tx.memCode, pcm, ivs.tx.rv, ivs.tx.ulIndex);
        ivs.tx.ulIndex++;
      }
      /* randomize transmit start */
      memcpy(ivs.tx.memDelay, &ivs.tx.memDelay[PCM_LENGTH], lpcm);
      memcpy(&ivs.tx.memDelay[PCM_LENGTH], pcm, lpcm);
      memcpy(pcm, &ivs.tx.memDelay[PCM_LENGTH - ivs.tx.ulDelay], lpcm);
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: IvsTxState                                               */
/*----------------------------------------------------------------------------*/
/* In:  Int16 msg       -> new control message symbol                         */
/*      Int16 metric    -> receiver metric (-1: ignore symbol)                */
/*      Bool  syncLock  -> indicates sync lock of control receiver            */
/*----------------------------------------------------------------------------*/

void IvsTxState(Int16 msg, Int16 metric, Bool syncLock) {
  Int16 k;

  if (syncLock == True) {
    ivs.tx.pendingStart = True;
  }
  if (msg >= SET_LLMSG + SET_HLMSG) {
    Abort("message symbol out of range, exit.");
  }
  else if (msg >= SET_LLMSG) {
    /* higher-layer message symbol */
    Ord8 dlHlackData = (Ord8)(msg - SET_LLMSG);

    if (metric < 0) {
      ivs.tx.stateIgn[msg]++;
      ivs.tx.stateIgn[DlMsgHlack]++;
    } else {
      ivs.tx.stateCnt[msg]++;
      ivs.tx.stateCnt[DlMsgHlack]++;
    }
    /* check conditions for successful HLACK reception */
    if (ivs.tx.state == IvsHlack) {
      LogInfo("[IVS ] ignoring additional HLACK (data: 0x0%X, metric: %3d)",
        dlHlackData, metric);
    }
    else if (
      ivs.tx.stateCnt[msg] >= 1 &&
      ivs.tx.stateCnt[msg]*2 + ivs.tx.stateIgn[msg] >= 4)
    {
      LogInfo("[IVS ] received HLACK (data: 0x0%X, metric: %3d), finished.",
        dlHlackData, metric);

      /* invoke callback functions */
      ivs.tx.state = IvsHlack;
      IvsReceiveHlack(dlHlackData);
      IvsCatchEvent(IVSEVENT_HLACKRECEIVED);

      /* entering new state -> reset all counters */
      for (k = 0; k <= SET_LLMSG + SET_HLMSG; k++) {
        ivs.tx.stateCnt[k] = 0;
        ivs.tx.stateIgn[k] = 0;
      }
      ivs.tx.stateCnt[DlMsgSend] = -1;
      ivs.tx.overallNack = 0;
    }
    else {
      LogInfo("[IVS ] received HLACK (data: 0x0%X, metric: %3d), waiting...",
        dlHlackData, metric);
    }
  }
  else {
    /* lower-layer message symbol */
    switch (msg) {

      case DlMsgStart:
        if (metric < 0 && ivs.tx.stateIgn[DlMsgStart] < IVS_GOSTART) {
          /* ignore, message reliability too low */
          ivs.tx.stateIgn[DlMsgStart]++;
          LogInfo("[IVS ] received START !> ignoring message (%s)",
            "reliability too low");
          break;
        }
        if (ivs.tx.pendingStart == True) {
          ivs.tx.pendingStart = False;
          ivs.tx.stateCnt[DlMsgStart] = LOCK_START_DL;
        } else {
          ivs.tx.stateCnt[DlMsgStart]++;
        }

        /* check start... */
        if (ivs.tx.stateCnt[DlMsgStart] > LOCK_START_DL + FAIL_RESTART &&
          (ivs.tx.stateCnt[DlMsgStart]-1)%FAIL_RESTART == 0)
        {
          LogInfo("[IVS ] received START (metric: %3d) too often, %s", metric,
            "starting over...");

          /* assume PSAP sync failure, starting over... */
          ivs.tx.ulDelay = (ivs.tx.ulDelay + IVS_TXINC)%PCM_LENGTH;
          ivs.tx.ulN     = NRF_SYNC + ivs.tx.mod.nfmuteall + ivs.tx.mod.nfdata;
          ivs.tx.ulIndex = 0;
          ivs.tx.rv      = 0;

          /* clear counter and invoke callback */
          ivs.tx.stateIgn[DlMsgStart] = 0;
          IvsCatchEvent(IVSEVENT_RESTARTMSD);
        }
        else if (ivs.tx.stateCnt[DlMsgStart] == LOCK_START_DL) {
          /* downlink sync locked, start sending */
          LogInfo("[IVS ] received START (metric: %3d), starting...", metric);

          /* choose type of uplink modulator */
          if (ivs.tx.overallNack < IVS_TXFAST) {
            SetModState(&ivs.tx.mod, Mod3bit4smp);
            LogInfo("[IVS ] fast modulator chosen (NACK count: %d)",
              ivs.tx.overallNack);
          }
          else {
            SetModState(&ivs.tx.mod, Mod3bit8smp);
            LogInfo("[IVS ] robust modulator chosen (NACK count: %d)",
              ivs.tx.overallNack);
          }
          /* define uplink parameters */
          ivs.tx.ulDelay = (ivs.tx.ulDelay + IVS_TXINC)%PCM_LENGTH;
          ivs.tx.ulN     = NRF_SYNC + ivs.tx.mod.nfmuteall + ivs.tx.mod.nfdata;
          ivs.tx.ulIndex = 0;
          ivs.tx.rv      = 0;

          /* set state and invoke callback */
          ivs.tx.state = IvsSendMsd;
          IvsCatchEvent(IVSEVENT_SENDINGMSD);

          /* entering new state -> reset other counters */
          ivs.tx.stateIgn[DlMsgStart] = 0;
          ivs.tx.stateCnt[DlMsgNack]  = 0;
          ivs.tx.stateCnt[DlMsgSend]  = 0;
        }
        else {
          /* preserve state */
          LogInfo("[IVS ] received START (metric: %3d)", metric);
        }
        if (ivs.tx.stateCnt[DlMsgAck] == 1) {
          /* START after single ACK -> assume ACK was incorrect */
          ivs.tx.stateCnt[DlMsgAck] = 0;
          LogInfo("[IVS ] ... !> assuming previous ACK was wrong");
        }
        if (ivs.tx.stateCnt[DlMsgHlack] + ivs.tx.stateIgn[DlMsgHlack] >= 1
          && ivs.tx.dlMsgOld < SET_LLMSG)
        {
          /* START after single HLACK -> assume HLACK was incorrect */
          for (k = SET_LLMSG; k <= SET_LLMSG + SET_HLMSG; k++) {
            ivs.tx.stateCnt[k] = 0;
            ivs.tx.stateIgn[k] = 0;
          }
          LogInfo("[IVS ] ... !> assuming previous HLACK was wrong");
        }
        break;

      case DlMsgNack:
        if (ivs.tx.pendingStart == True ||
          ivs.tx.state == IvsAck || ivs.tx.state == IvsHlack)
        {
          /* downlink sync locked, still waiting to start */
          LogInfo("[IVS ] received NACK  !> ignoring message (%s)",
            "expecting START");
          break;
        }
        else if (metric < 0) {
          /* ignore, message reliability too low */
          LogInfo("[IVS ] received NACK  !> ignoring message (%s)",
            "reliability too low");
          break;
        }
        /* message seems reliable, increase counters */
        ivs.tx.stateCnt[DlMsgNack]++;
        ivs.tx.overallNack++;

        if (
          ivs.tx.stateCnt[DlMsgNack] == 1 &&
          ivs.tx.stateCnt[DlMsgSend] < 0)
        {
          /* ignore, NACK should only follow START or NACK */
          ivs.tx.stateCnt[DlMsgNack] = 0;
          LogInfo("[IVS ] received NACK  !> ignoring message (%s)",
            "assume IDLE");
        }
        else if (ivs.tx.stateCnt[DlMsgNack] == 1) {
          LogInfo("[IVS ] received NACK  (metric: %3d), first time", metric);

          /* entering new state -> reset other counters */
          ivs.tx.stateIgn[DlMsgStart] = 0;
          ivs.tx.stateCnt[DlMsgStart] = 0;
          ivs.tx.stateCnt[DlMsgSend]  = 0;
        }
        else {
          /* preserve state */
          LogInfo("[IVS ] received NACK  (metric: %3d)", metric);
        }
        if (ivs.tx.stateCnt[DlMsgAck] == 1) {
          /* NACK after single ACK -> assume ACK was incorrect */
          ivs.tx.stateCnt[DlMsgAck] = 0;
          LogInfo("[IVS ] ... !> assuming previous ACK was wrong");
        }
        if (ivs.tx.stateCnt[DlMsgHlack] + ivs.tx.stateIgn[DlMsgHlack] >= 1
          && ivs.tx.dlMsgOld < SET_LLMSG)
        {
          /* NACK after single HLACK -> assume HLACK was incorrect */
          for (k = SET_LLMSG; k <= SET_LLMSG + SET_HLMSG; k++) {
            ivs.tx.stateCnt[k] = 0;
            ivs.tx.stateIgn[k] = 0;
          }
          LogInfo("[IVS ] ... !> assuming previous HLACK was wrong");
        }
        break;

      case DlMsgAck:
        if (ivs.tx.pendingStart == True || ivs.tx.state == IvsHlack) {
          /* downlink sync locked, still waiting to start */
          LogInfo("[IVS ] received ACK   !> ignoring message (%s)",
            "expecting START");
          break;
        }
        else if (ivs.tx.state == IvsAck) {
          /* ACK already recorded */
          LogInfo("[IVS ] received ACK   !> ignoring message (%s)",
            "already recorded");
          break;
        }
        else if (metric < 0) {
          /* ignore, message reliability too low */
          LogInfo("[IVS ] received ACK   !> ignoring message (%s)",
            "reliability too low");
          break;
        }
        /* message seems reliable, two ACK messages required */
        ivs.tx.stateCnt[DlMsgAck]++;

        if (
          ivs.tx.stateCnt[DlMsgAck] == 1 &&
          ivs.tx.stateCnt[DlMsgSend] < 0)
        {
          /* ignore, IDLE mode precedes ACK */
          ivs.tx.stateCnt[DlMsgAck] = 0;
          LogInfo("[IVS ] received ACK   !> ignoring message (%s)",
            "assume IDLE");
        }
        else if (ivs.tx.stateCnt[DlMsgAck] == 2) {
          LogInfo("[IVS ] received ACK   (metric: %3d), %s", metric,
            "ACK recorded, wait for HLACK");

          /* invoke callback functions and wait for HLACK */
          ivs.tx.state = IvsAck;
          IvsReceiveAck();
          IvsCatchEvent(IVSEVENT_LLACKRECEIVED);

          /* entering new state -> reset all counters */
          for (k = 0; k < SET_LLMSG; k++) {
            ivs.tx.stateCnt[k] = 0;
            ivs.tx.stateIgn[k] = 0;
          }
          ivs.tx.stateCnt[DlMsgSend] = -1;
          ivs.tx.overallNack = 0;
        }
        else {
          /* waiting... */
          LogInfo("[IVS ] received ACK   (metric: %3d), waiting...", metric);
        }
        if (ivs.tx.stateCnt[DlMsgHlack] + ivs.tx.stateIgn[DlMsgHlack] >= 1
          && ivs.tx.dlMsgOld < SET_LLMSG)
        {
          /* ACK after single HLACK -> assume HLACK was incorrect */
          for (k = SET_LLMSG; k <= SET_LLMSG + SET_HLMSG; k++) {
            ivs.tx.stateCnt[k] = 0;
            ivs.tx.stateIgn[k] = 0;
          }
          LogInfo("[IVS ] ... !> assuming previous HLACK was wrong");
        }
        break;

      default:
        if (ivs.tx.pendingStart == True ||
          ivs.tx.state == IvsAck || ivs.tx.state == IvsHlack)
        {
          /* downlink sync locked, still waiting to start */
          LogInfo("[IVS ] received IDLE  !> ignoring message (%s)",
            "expecting START");
        }
        else if (
          ivs.tx.stateCnt[DlMsgStart] > 0 ||
          ivs.tx.stateCnt[DlMsgNack] > 0 ||
          ivs.tx.stateCnt[DlMsgAck] > 0)
        {
          LogInfo("[IVS ] received IDLE  !> ignoring message");
        }
        else {
          LogInfo("[IVS ] remaining in IDLE mode");
        }
        break;
    }
  }
  /* store message if reliable */
  if (metric > 0) {
    ivs.tx.dlMsgOld = msg;
  }
}
