/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         modem_psap.c                                                 */
/* Version:      8.6.0 (Rel8) / 9.4.0 (Rel9)                                  */
/* Date:         2011-02-08                                                   */
/* Description:  eCall PSAP modem                                             */
/*----------------------------------------------------------------------------*/

#include "modemx.h"


static PsapData psap;

/* PSAP external reference */
PsapData* PsapGetData(void) { return &psap; }

/*============================================================================*/
/* PSAP IMPLEMENTATION: PsapReset, PsapSendStart, PsapSendHlack               */
/*----------------------------------------------------------------------------*/

void PsapReset(void) {
  PsapRxReset();
  PsapTxReset();
}

void PsapSendStart(void) {
  if (psap.rx.state == PsapIdle) {
    /* START trigger out of IDLE state */
    LogInfo("[PSAP] SendStart received from control");
    LogInfo("[----]");
    PsapRxReset();

    psap.rx.state = PsapTrigger;
    PsapCatchEvent(PSAPEVENT_SENDINGSTART);
  }
  else if (psap.rx.state == PsapNack) {
    /* START trigger out of NACK state */
    LogInfo("[PSAP] SendStart received, starting over...");
    LogInfo("[----]");
    SyncReset(&psap.rx.sync, NULL, NULL);

    psap.rx.state      = PsapTrigger;
    psap.rx.dlMsgCnt   = -1;
    psap.rx.rv         = 0;
    psap.rx.mgIndex    = 0;
    psap.rx.ulIndex    = 0;
    psap.rx.ulSyncTail = 0;

    PsapCatchEvent(PSAPEVENT_SENDINGRESTART);
  }
  else {
    /* ignore otherwise */
    LogInfo("[PSAP] SendStart received from control, ignored.");
    LogInfo("[----]");
  }
}

void PsapSendHlack(const Ord8 data) {
  if (psap.rx.state == PsapAck) {
    /* HLACK mode requires MSD to be successfully received */
    LogInfo("[PSAP] HLACK data received from control");
    LogInfo("[----]");

    psap.rx.state       = PsapHlack;
    psap.rx.dlMsgCnt    = -1;
    psap.rx.dlHlackData = data & 0x0F;

    PsapCatchEvent(PSAPEVENT_SENDINGHLACK);
  }
  else {
    /* ignore otherwise */
    LogInfo("[PSAP] HLACK data received from control, ignored.");
    LogInfo("[----]");
  }
}

/*----------------------------------------------------------------------------*/

void PsapRxReset(void) {
  CtrlRxData *ctrl = &psap.rx.ctrl;
  SyncState  *sync = &psap.rx.sync;

  /* reset state */
  psap.rx.state = PsapIdle;

  /* assign buffer pointers */
  psap.rx.msdBin  = (Ord1*)  &psap.rx.buffer[0];
  psap.rx.msd     = (Ord8*)  &psap.rx.buffer[NRB_INFO_CRC];
  psap.rx.memCtrl = (Int16*) &psap.rx.buffer[sizeof(IntLLR)*NRB_CODE_ARQ];
  psap.rx.memCode = (IntLLR*)&psap.rx.buffer[0];

  /* reset control receiver */
  CtrlRxReset(ctrl, PortPsap, sync, psap.rx.memCtrl, LOCK_START_UL);

  /* reset uplink counters */
  psap.rx.rv         = 0;
  psap.rx.ulIndex    = 0;
  psap.rx.mgIndex    = 0;
  psap.rx.ulTrials   = 0;
  psap.rx.ulSyncTail = 0;

  /* reset HLACK data and downlink counters */
  psap.rx.dlHlackData = 0;
  psap.rx.dlData      = DlMsgNoop;
  psap.rx.dlMsgCnt    = -1;
  psap.rx.dlIndex     = -1;

  /* reset sync */
  SyncReset(sync,
    (Int32*)&psap.rx.buffer[0
    + sizeof(IntLLR)* NRB_CODE_ARQ
    + sizeof(Int16) * NRF_MEMCTRL*PCM_LENGTH],
    (Int32*)&psap.rx.buffer[0
    + sizeof(IntLLR)* NRB_CODE_ARQ
    + sizeof(Int16) * NRF_MEMCTRL*PCM_LENGTH
    + sizeof(Int32) * NRS_MEMSYNC]);

  /* clear buffer */
  memset(psap.rx.buffer, 0, sizeof(char)*(0
    + sizeof(IntLLR)* NRB_CODE_ARQ
    + sizeof(Int16) * NRF_MEMCTRL*PCM_LENGTH
    + sizeof(Int32) * NRS_MEMSYNC
    + sizeof(Int32) * 2*(NRF_SYNC+1)));

  LogInfo("[PSAP] reset receiver");
}

/*----------------------------------------------------------------------------*/

void PsapTxReset(void) {
  CtrlTxReset(&psap.tx.ctrl, PortPsap);

  LogInfo("[PSAP] reset transmitter");
}

/*============================================================================*/
/* PSAP IMPLEMENTATION: PsapProcess                                           */
/*----------------------------------------------------------------------------*/
/* InOut:  Int16* pcm  <-> input and output frame of 16bit PCM samples        */
/*----------------------------------------------------------------------------*/

void PsapProcess(Int16 *pcm) {
  Int16 i;
  CtrlPort *prx = &psap.rx.ctrl.port;
  CtrlPort *ptx = &psap.tx.ctrl.port;

  /* invert input signal if required */
  if (prx->invert == True) {
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] = (pcm[i] > PCM_MIN) ? -pcm[i] : PCM_MAX;
    }
  }
  PsapRxProcess(pcm);

  if (prx->state == CtrlTxIdle || prx->state == CtrlTxSend) {
    /* port match, interface copy */
    ptx->state        = prx->state;
    ptx->u.tx.dlData  = prx->u.tx.dlData;
    ptx->u.tx.dlIndex = prx->u.tx.dlIndex;
  }
  else {
    /* port mismatch, no downlink payload */
    ptx->state        = CtrlTxIdle;
    ptx->u.tx.dlData  = DlMsgNoop;
    ptx->u.tx.dlIndex = -1;
  }
  PsapTxProcess(pcm);

  /* invert output signal if required */
  if (ptx->invert == True) {
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] = (pcm[i] > PCM_MIN) ? -pcm[i] : PCM_MAX;
    }
  }
}

/*----------------------------------------------------------------------------*/

void PsapRxProcess(const Int16 *pcm) {
  Int16 i;

  /* PSAP receiver, copy PCM data to input buffer */
  for (i = 0; i < NRF_MEMCTRL-1; i++) {
    memcpy(&psap.rx.memCtrl[i*PCM_LENGTH],
      &psap.rx.memCtrl[(i+1)*PCM_LENGTH], sizeof(Int16)*PCM_LENGTH);
  }
  memcpy(&psap.rx.memCtrl[(NRF_MEMCTRL-1)*PCM_LENGTH], pcm,
    sizeof(Int16)*PCM_LENGTH);

  /* process uplink and downlink state machine */
  PsapRxUplink(pcm);
  PsapRxDownlink();
}

/*----------------------------------------------------------------------------*/

void PsapTxProcess(Int16 *pcm) {
  CtrlTxProcess(&psap.tx.ctrl, pcm);
}

/*============================================================================*/
/* UTILITY FUNCTION: PsapRxUplink                                             */
/*----------------------------------------------------------------------------*/
/* In:  const Int16* pcm  -> input frame of 16bit PCM samples                 */
/*----------------------------------------------------------------------------*/

void PsapRxUplink(const Int16 *pcm) {
  Int16 temp;
  Bool  crc, decode;

  CtrlRxData *ctrl = &psap.rx.ctrl;
  SyncState  *sync = &psap.rx.sync;
  ModState   *mod  = &psap.rx.mod;

  /* process each state individually */
  switch (psap.rx.state) {
    case PsapIdle:
      /* run control receiver and watch for SEND signal */
      CtrlRxProcess(ctrl, pcm);

      if (ctrl->port.u.rx.dlData == DlMsgSend) {
        LogInfo("[PSAP] received SEND  (metric: %3d), %s",
          ctrl->port.u.rx.dlMetric, "initiating START trigger");

        /* reset receiver and prepare for MSD reception */
        PsapRxReset();
        psap.rx.state = PsapTrigger;
        PsapCatchEvent(PSAPEVENT_SENDINGSTART);
      }
      else if (ctrl->port.u.rx.dlData == DlMsgReset) {
        PsapReset();
        PsapCatchEvent(PSAPEVENT_IDLEPOSTRESET);
      }
      break;

    case PsapTrigger:
      /* start sending control messages, wait for IVS to transmit */
      ctrl->port.state = CtrlTxIdle;
      break;

    case PsapStart:
      /* run sync with new input data */
      ctrl->port.state = CtrlTxIdle;
      Sync(sync, pcm, "PSAP", ctrl->port.invert);

      if (sync->flag == True && sync->delay <= 0) {
        /* sync successful */
        ctrl->port.state = CtrlTxSend;
        PsapCatchEvent(PSAPEVENT_MSDSYNC);

        /* apply sync inversion flag */
        if ((ctrl->port.invert = sync->invert) == True) {
          LogInfo("[PSAP] codec inversion detected!");
        }
        /* start sending NACK messages */
        psap.rx.state    = PsapNack;
        psap.rx.dlMsgCnt = -1;

        /* store amplitude and choose modulator */
        sync->amplitude[2] = sync->amplitude[0];
        PsapRxSetMod(sync, mod);

        /* check for muting gaps passed already */
        if (sync->delay <= -PCM_LENGTH) {
          temp = MIN(-(sync->delay)/PCM_LENGTH, mod->nfmute1);
          sync->delay += temp*PCM_LENGTH;
          psap.rx.mgIndex += temp;
        }
      }
      break;

    case PsapNack:
      /* check for better sync for some more frames */
      if (psap.rx.ulSyncTail < NRF_OBSERVE) {
        psap.rx.ulSyncTail++;

        /* run synchronization */
        Sync(sync, pcm, "PSAP", ctrl->port.invert);

        if (sync->flag == True && sync->delay <= 0) {
          sync->amplitude[2] = sync->amplitude[0];

          /* apply sync inversion flag */
          if ((ctrl->port.invert = sync->invert) == True) {
            LogInfo("[PSAP] codec inversion detected (sync observer)!");
          }
          /* reset receiver, previous sync was likely too early */
          psap.rx.rv         = 0;
          psap.rx.ulIndex    = 0;
          psap.rx.mgIndex    = -1;
          psap.rx.ulSyncTail = 0;

          /* reinitialize buffer */
          memset(psap.rx.memCode, 0, sizeof(IntLLR)*NRB_CODE_ARQ);

          /* choose modulator */
          PsapRxSetMod(sync, mod);

          /* check for muting gaps passed already */
          if (sync->delay <= -PCM_LENGTH) {
            temp = MIN(-(sync->delay)/PCM_LENGTH, mod->nfmute1);
            sync->delay += temp*PCM_LENGTH;
            psap.rx.mgIndex += temp;
          }
        }
      } else {
        /* continue updating sync filter state for sync tracking purposes */
        SyncFilter(sync, pcm, ctrl->port.invert);
      }
      /* run demodulation if not muting */
      decode = False;

      if (psap.rx.mgIndex >= 0) {
        if (mod->mgTable[psap.rx.mgIndex] >= 0) {
          PsapReceiver(mod,
            &psap.rx.memCtrl[(NRF_MEMCTRL-1)*PCM_LENGTH + sync->delay],
            &psap.rx.memCode[(mod->mfpf*mod->bpsym)*psap.rx.ulIndex]);
          psap.rx.ulIndex++;
          decode = True;
        }
        else if (mod->mgTable[psap.rx.mgIndex] == -6) {
          /* track sync peaks and correct delay if necessary */
          SyncTrack(sync, ctrl->port.invert);

          if (sync->flag == True) {
            while (sync->delay > 0) {
              sync->delay -= PCM_LENGTH;
              psap.rx.mgIndex--;
            }
            while (sync->delay <= -PCM_LENGTH) {
              sync->delay += PCM_LENGTH;
              psap.rx.mgIndex++;

              if (mod->mgTable[psap.rx.mgIndex] >= 0) {
                PsapReceiver(mod,
                  &psap.rx.memCtrl[(NRF_MEMCTRL-1)*PCM_LENGTH + sync->delay],
                  &psap.rx.memCode[(mod->mfpf*mod->bpsym)*psap.rx.ulIndex]);
                psap.rx.ulIndex++;
              }
            }
          }
          if (sync->checkCnt >= SYNC_BADTRACK) {
            /* too many tracking failures */
            LogInfo("[PSAP] restart due to tracking failures!");
            LogInfo("[----]");
            PsapSendStart();
            break;
          }
        }
      }
      /* increase muting gap counter */
      psap.rx.mgIndex++;

      /* run decoder if desired */
      if (decode == True && (
        ((psap.rx.rv%ARQ_MAX) > 0 && psap.rx.ulIndex == mod->decpos1) ||
        ((psap.rx.rv%ARQ_MAX) > 0 && psap.rx.ulIndex == mod->decpos2) ||
        psap.rx.ulIndex == psap.rx.ulN))
      {
        /* do not run twice with rv = 0, buffer will be reinitialized */
        crc = FecDecode(psap.rx.memCode, psap.rx.rv, psap.rx.msdBin);

        if (crc == False) {
          /* CRC failed, initialization for new redundancy version */
          LogInfo("[PSAP] CRC failed; rv = %d", psap.rx.rv);
          LogInfo("[----]");

          /* reinitialize buffer */
          memset(psap.rx.memCode, 0, sizeof(IntLLR)*NRB_CODE_ARQ);

          /* full block was received, reinitialization necessary */
          if (psap.rx.ulIndex == psap.rx.ulN) {
            psap.rx.ulIndex = 0;
            psap.rx.rv++;

            if ((psap.rx.rv%ARQ_MAX) == 0) {
              /* too many retransmissions, sync must have failed */
              PsapSendStart();
              psap.rx.ulTrials++;
            } else
              psap.rx.mgIndex = -(mod->nfmute4);
          }
        } else {
          LogInfo("[PSAP] MSD successfully received; %s: %d",
            "redundancy versions", psap.rx.rv);

          /* set ACK state prior to any MSD callback */
          psap.rx.state    = PsapAck;
          psap.rx.dlMsgCnt = -1;

          /* inform application of received MSD */
          Bit2Byte(psap.rx.msdBin, psap.rx.msd, MSD_MAX_LENGTH);
          PsapReceiveMsd(psap.rx.msd, MSD_MAX_LENGTH);
          PsapCatchEvent(PSAPEVENT_MSDRECEIVED);

          /* MSD callback may have triggered HLACK immediately */
          if (psap.rx.state == PsapAck) {
            PsapCatchEvent(PSAPEVENT_SENDINGLLACK);
          }
        }
      }
      break;

    case PsapAck:
      /* MSD was received, keep sending ACK messages */
      break;
    case PsapHlack:
      /* MSD was received, higher-layer message in progress */
      break;
    default:
      /* error */
      Abort("PSAP state not recognized, exit.");
      break;
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: PsapRxDownlink                                           */
/*----------------------------------------------------------------------------*/

void PsapRxDownlink(void) {
  CtrlPort *port = &psap.rx.ctrl.port;

  if (
    port->state == CtrlRxIdle ||
    port->state == CtrlRxSync ||
    port->state == CtrlRxLock)
  {
    /* PSAP in control receiver mode, port data was set above */
  }
  else if (psap.rx.state == PsapIdle && psap.rx.dlMsgCnt >= 0) {
    /* final message has been sent, stop output */
    port->u.tx.dlData  = DlMsgNoop;
    port->u.tx.dlIndex = -1;
  }
  else {
    /* interface copy */
    port->u.tx.dlData  = psap.rx.dlData;
    port->u.tx.dlIndex = psap.rx.dlIndex;

    /* increase downlink index */
    psap.rx.dlIndex++;
    psap.rx.dlIndex %= NRF_DLCHUNK;

    if (psap.rx.dlIndex == 0) {
      /* negative counter indicates new message */
      if (psap.rx.dlMsgCnt < 0) {
        psap.rx.dlMsgCnt = 0;
      } else {
        psap.rx.dlMsgCnt++;
      }
      /* check message counter and change state */
      switch (psap.rx.state) {
        case PsapTrigger:
          if (psap.rx.dlMsgCnt >= LOCK_START_DL) {
            psap.rx.state    = PsapStart;
            psap.rx.dlMsgCnt = 0;
          }
          /* set message right here to trigger sync preamble output */
          psap.rx.dlData = DlMsgStart;
          break;
        case PsapStart:
          if (psap.rx.dlMsgCnt >= (PSAP_NUMSTART - LOCK_START_DL)) {
            psap.rx.state    = PsapIdle;
            psap.rx.dlMsgCnt = 0;
            PsapCatchEvent(PSAPEVENT_IDLEPOSTSTART);
          }
          break;
        case PsapAck:
          if (psap.rx.dlMsgCnt >= PSAP_NUMACK) {
            psap.rx.state    = PsapIdle;
            psap.rx.dlMsgCnt = 0;
            PsapCatchEvent(PSAPEVENT_IDLEPOSTLLACK);
          }
          break;
        case PsapHlack:
          if (psap.rx.dlMsgCnt >= PSAP_NUMHLACK) {
            psap.rx.state    = PsapIdle;
            psap.rx.dlMsgCnt = 0;
            PsapCatchEvent(PSAPEVENT_IDLEPOSTHLACK);
          } else {
            LogInfo("[PSAP] sending  HLACK; data: 0x0%X", psap.rx.dlHlackData);

            /* add message offset */
            psap.rx.dlData = SET_LLMSG + psap.rx.dlHlackData;
          }
          break;
        default:
          /* other states do not require message counting */
          break;
      }
    }
    else if (psap.rx.dlIndex == NRF_SYNC + NRF_DLMUTE1LL) {
      /* start of downlink data portion, set next message */
      switch (psap.rx.state) {
        case PsapTrigger:
        case PsapStart:
          psap.rx.dlData = DlMsgStart;
          if (psap.rx.dlMsgCnt < 0) {
            psap.rx.dlMsgCnt = 0;
          }
          LogInfo("[PSAP] sending  START");
          break;
        case PsapNack:
          psap.rx.dlData = DlMsgNack;
          if (psap.rx.dlMsgCnt < 0) {
            psap.rx.dlMsgCnt = 0;
          }
          LogInfo("[PSAP] sending  NACK");
          break;
        case PsapAck:
          psap.rx.dlData = DlMsgAck;
          if (psap.rx.dlMsgCnt < 0) {
            psap.rx.dlMsgCnt = 0;
          }
          LogInfo("[PSAP] sending  ACK");
          break;
        case PsapHlack:
          if (psap.rx.dlMsgCnt < 0) {
            psap.rx.dlData = DlMsgAck;
            LogInfo("[PSAP] sending  ACK   (HLACK pending)");
          }
          break;
        default:
          LogInfo("[PSAP] remaining in IDLE mode");
          break;
      }
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: PsapRxSetMod                                             */
/*----------------------------------------------------------------------------*/
/* In:  const SyncState* sync  -> sync struct                                 */
/* Out: ModState*        mod   -> modulator struct                            */
/*----------------------------------------------------------------------------*/

void PsapRxSetMod(const SyncState *sync, ModState *mod)
{
  Int32 xCorrA =
    sync->memWakeup[0] +
    sync->memWakeup[2] +
    sync->memWakeup[4] +
    sync->memWakeup[6];
  Int32 xCorrB =
    sync->memWakeup[1] +
    sync->memWakeup[3] +
    sync->memWakeup[5] +
    sync->memWakeup[7];

  if (xCorrA/MAX(xCorrB, 1) > PSAP_THRESHOLD) {
    SetModState(mod, Mod3bit4smp);
    LogInfo("[PSAP] fast modulator chosen (metric: %d)",
      xCorrA/MAX(xCorrB, 1));
  }
  else if (xCorrB/MAX(xCorrA, 1) > PSAP_THRESHOLD) {
    SetModState(mod, Mod3bit8smp);
    LogInfo("[PSAP] robust modulator chosen (metric: %d)",
      xCorrB/MAX(xCorrA, 1));
  }
  else if (psap.rx.ulTrials == 0) {
    SetModState(mod, Mod3bit4smp);
    LogInfo("[PSAP] fast modulator chosen (metric: %d)",
      xCorrA/MAX(xCorrB, 1));
  }
  else {
    SetModState(mod, Mod3bit8smp);
    LogInfo("[PSAP] robust modulator chosen (metric: %d)",
      xCorrB/MAX(xCorrA, 1));
  }
  psap.rx.ulN = mod->nfdata;
}
