/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         standalone.c                                                 */
/* Version:      8.6.0 (Rel8) / 9.4.0 (Rel9)                                  */
/* Date:         2011-02-08                                                   */
/* Description:  eCall standalone wrapper                                     */
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "modemx.h"
#include "standalone.h"


static EcallShell ecs;

/*============================================================================*/
/* Callback functions                                                         */
/*----------------------------------------------------------------------------*/

void IvsCatchEvent(IvsEvent ie) {
  switch (ie) {
    /* custom event handlers to be implemented here... */
    /* see modemx.h for a list of possible events (enum IvsEvent) */
    default:
      break;
  }
}

void IvsReceiveAck(void) {
  LogInfo("[CTRL] ACK received from IVS");
  LogInfo("[----]");
}

void IvsReceiveHlack(const Ord8 data) {
  if ((ecs.msd[0] & 0x0F) == (data & 0x0F)) {
    LogInfo("[CTRL] HLACK data received from IVS; correct.");
    LogInfo("[====]");
  } else {
    LogInfo("[CTRL] HLACK data received from IVS; mismatch!");
    LogInfo("[====]");
  }
}

void PsapCatchEvent(PsapEvent pe) {
  switch (pe) {
    /* custom event handlers to be implemented here... */
    /* see modemx.h for a list of possible events (enum PsapEvent) */
    default:
      break;
  }
}

void PsapReceiveMsd(const Ord8 *msd, int length) {
  int n;
  for (n = 0; n < MSD_MAX_LENGTH; n++) {
    if (ecs.msd[n] != msd[n]) break;
  }
  if (n < MSD_MAX_LENGTH) {
    LogInfo("[CTRL] MSD received from PSAP; mismatch!");
    LogInfo("[====]");
  } else {
    LogInfo("[CTRL] MSD received from PSAP; correct.");
    LogInfo("[====]");
  }
  ecs.msdReceived = True;
}

void LogInfo(const char *format, ...) {
  va_list args;

  va_start(args, format);
  fprintf (stdout, "INFO  ");
  vfprintf(stdout, format, args);
  fprintf (stdout, "\n");
  va_end  (args);
  fflush  (stdout);
}

void Abort(const char *format, ...) {
  va_list args;

  va_start(args, format);
  fprintf (stderr, "ABORT ");
  vfprintf(stderr, format, args);
  fprintf (stderr, "\n");
  va_end  (args);
  fflush  (stderr);

  exit(1);
}

/*============================================================================*/
/* main() wrapper                                                             */
/*----------------------------------------------------------------------------*/

int main(int argc, char *argv[]) {
  int  n;
  char msdChar[501];
  FILE *campfile;

  /* Argument parsing */
  Parse(argc, argv);

  /* Open campaign file */
  if ((campfile = fopen(ecs.campaign, "r")) == NULL) {
    Abort("unable to open campaign file %s", ecs.campaign);
  }
  for (;;) {
    n = fscanf(campfile,
      "%d %500s %*10s %*10s %*10s %*d %*d %*100s %*d %*d %*d %*100s %*d\n",
      &ecs.indexCase, msdChar);

    /* Check for end of file and successful read */
    if (n <= 0 && feof(campfile)) {
      break;
    } else if (n != 2) {
      Abort("unable to read enough data from campaign file.");
    }
    printf("running test case: %d\n", ecs.indexCase);

    /* Convert MSD to byte representation */
    ConvertMsd(msdChar, ecs.msd);

    /* Reset the eCall modem and process the test case */
    EcallReset(ecs.indexCase);
    EcallProcess();
  }
  /* Close file */
  if (campfile != NULL) fclose(campfile);
  return 0;
}

/*============================================================================*/
/* EcallReset()                                                               */
/*----------------------------------------------------------------------------*/

void EcallReset(int index) {
  char prxfile[501];
  char ptxfile[501];

  switch (ecs.mode) {
    case Ivs:
      if (ecs.portbase != NULL) {
        sprintf(prxfile, "%sivsrx%d.txt", ecs.portbase, index);
        sprintf(ptxfile, "%sivstx%d.txt", ecs.portbase, index);
        PortLogInit(&ecs.prxlog, prxfile);
        PortLogInit(&ecs.ptxlog, ptxfile);
      }
      IvsReset(ecs.msd, MSD_MAX_LENGTH);
      break;
    case IvsRx:
      if (ecs.outbase != NULL) {
        sprintf(prxfile, "%sivsrx%d.txt", ecs.outbase, index);
        PortLogInit(&ecs.prxlog, prxfile);
      } else {
        PortLogInit(&ecs.prxlog, NULL);
      }
      IvsRxReset();
      break;
    case IvsTx:
      if (ecs.inbase != NULL) {
        sprintf(ptxfile, "%sivstx%d.txt", ecs.inbase, index);
        PortLogInit(&ecs.ptxlog, ptxfile);
      } else {
        PortLogInit(&ecs.ptxlog, NULL);
      }
      IvsTxReset(ecs.msd, MSD_MAX_LENGTH);
      break;
    case Psap:
      if (ecs.portbase != NULL) {
        sprintf(prxfile, "%spsaprx%d.txt", ecs.portbase, index);
        sprintf(ptxfile, "%spsaptx%d.txt", ecs.portbase, index);
        PortLogInit(&ecs.prxlog, prxfile);
        PortLogInit(&ecs.ptxlog, ptxfile);
      }
      PsapReset();
      ecs.msdReceived = False;
      break;
    case PsapRx:
      if (ecs.outbase != NULL) {
        sprintf(prxfile, "%spsaprx%d.txt", ecs.outbase, index);
        PortLogInit(&ecs.prxlog, prxfile);
      } else {
        PortLogInit(&ecs.prxlog, NULL);
      }
      PsapRxReset();
      ecs.msdReceived = False;
      break;
    case PsapTx:
      if (ecs.inbase != NULL) {
        sprintf(ptxfile, "%spsaptx%d.txt", ecs.inbase, index);
        PortLogInit(&ecs.ptxlog, ptxfile);
      } else {
        PortLogInit(&ecs.ptxlog, NULL);
      }
      PsapTxReset();
      ecs.msdReceived = False;
      break;
    default:
      Abort("execution mode not recognized.");
      break;
  }
}

/*============================================================================*/
/* EcallProcess()                                                             */
/*----------------------------------------------------------------------------*/

void EcallProcess(void) {
  int n, m;

  /* Get pointers to modem data structs */
  IvsData  *ivs  = IvsGetData();
  PsapData *psap = PsapGetData();

  /* Clear buffer */
  memset(ecs.pcmBuffer, 0, sizeof(*ecs.pcmBuffer)*MAX_FRAMES*PCM_LENGTH);

  /* Get PCM frames into buffer */
  if (ecs.mode == Ivs || ecs.mode == IvsRx) {
    if ((ecs.frameCount = GetBuffer(ecs.inbase, "dlout", ecs.indexCase, 1))
      <= ecs.frameStart)
    { Abort("PCM data contains idle frames only!"); }
  }
  else if (ecs.mode == Psap || ecs.mode == PsapRx) {
    if ((ecs.frameCount = GetBuffer(ecs.inbase, "ulout", ecs.indexCase, 1))
      <= ecs.frameStart)
    { Abort("PCM data contains idle frames only!"); }
  }

  /* Modem processing depending on run mode */
  switch (ecs.mode) {
    case Ivs:
      for (n = 0; n < ecs.frameCount; n++) {
        if (ecs.ivsPush == True && n == ecs.frameStart) {
          IvsSendStart();
        }
        IvsProcess(&ecs.pcmBuffer[n*PCM_LENGTH]);
        if (ecs.portbase != NULL) {
          PortLogWrite(&ecs.prxlog, &ivs->rx.ctrl.port);
          PortLogWrite(&ecs.ptxlog, &ivs->tx.ctrl.port);
        }
      }
      break;
    case IvsRx:
      for (n = 0; n < ecs.frameCount; n++) {
        if (ecs.ivsPush == True && n == ecs.frameStart) {
          IvsSendStart();
        }
        IvsRxProcess(&ecs.pcmBuffer[n*PCM_LENGTH]);
        PortLogWrite(&ecs.prxlog, &ivs->rx.ctrl.port);
      }
      break;
    case IvsTx:
      ecs.frameCount = 0;
      while ((n = PortLogRead(&ecs.ptxlog, &ivs->tx.ctrl.port)) > 0) {
        IvsTxProcess(&ecs.pcmBuffer[ecs.frameCount*PCM_LENGTH]);
        ecs.frameCount++;
      }
      if (n < 0) {
        Abort("unable to read from port file %s", ecs.ptxlog.filename);
      }
      break;
    case Psap:
      for (n = 0, m = 0; n < ecs.frameCount; n++) {
        if (ecs.ivsPush == False && n == ecs.frameStart) {
          PsapSendStart();
        }
        if (ecs.msdReceived == True) {
          if (++m == ecs.frameStartHlack)
            PsapSendHlack(ecs.msd[0] & 0x0F);
        }
        PsapProcess(&ecs.pcmBuffer[n*PCM_LENGTH]);
        if (ecs.portbase != NULL) {
          PortLogWrite(&ecs.prxlog, &psap->rx.ctrl.port);
          PortLogWrite(&ecs.ptxlog, &psap->tx.ctrl.port);
        }
      }
      break;
    case PsapRx:
      for (n = 0, m = 0; n < ecs.frameCount; n++) {
        if (ecs.ivsPush == False && n == ecs.frameStart) {
          PsapSendStart();
        }
        if (ecs.msdReceived == True) {
          if (++m == ecs.frameStartHlack)
            PsapSendHlack(ecs.msd[0] & 0x0F);
        }
        PsapRxProcess(&ecs.pcmBuffer[n*PCM_LENGTH]);
        PortLogWrite(&ecs.prxlog, &psap->rx.ctrl.port);
      }
      break;
    case PsapTx:
      ecs.frameCount = 0;
      while ((n = PortLogRead(&ecs.ptxlog, &psap->tx.ctrl.port)) > 0) {
        PsapTxProcess(&ecs.pcmBuffer[ecs.frameCount*PCM_LENGTH]);
        ecs.frameCount++;
      }
      if (n < 0) {
        Abort("unable to read from port file %s", ecs.ptxlog.filename);
      }
      break;
    default:
      Abort("execution mode not recognized.");
      break;
  }

  /* Write PCM output to file if specified */
  if (ecs.outbase != NULL) {
    if (ecs.mode == Ivs || ecs.mode == IvsTx) {
      if (PutBuffer(ecs.outbase, "ulin", ecs.indexCase, ecs.frameCount)
        < ecs.frameCount)
      { Abort("writing to file failed."); }
    }
    else if (ecs.mode == Psap || ecs.mode == PsapTx) {
      if (PutBuffer(ecs.outbase, "dlin", ecs.indexCase, ecs.frameCount)
        < ecs.frameCount)
      { Abort("writing to file failed."); }
    }
  }
}

/*============================================================================*/
/* Utility functions                                                          */
/*----------------------------------------------------------------------------*/

void Parse(int argc, char **argv) {
  int n;
  const char modeName[NUM_MODES][7] = {
    "ivs", "ivsrx", "ivstx", "psap", "psaprx", "psaptx"
  };
  ecs.mode            = Udef;
  ecs.ivsPush         = False;
  ecs.msdReceived     = False;
  ecs.campaign        = NULL;
  ecs.inbase          = NULL;
  ecs.outbase         = NULL;
  ecs.portbase        = NULL;
  ecs.frameStart      = 50;
  ecs.frameStartHlack = 10;

  while ((--argc>0) && ((*++argv)[0]=='-')) {
    switch ((*argv)[1]) {
      case 'm':
        if ((--argc>0) && ((*++argv)[0]!='-')) {
          for (n = 0; n < NUM_MODES; n++) {
            if (strcmp(*argv, modeName[n]) == 0) {
              ecs.mode = n;
              break;
            }
          }
          if (ecs.mode == Udef) { PrintUsage("-m invalid argument"); }
        }
        else { PrintUsage("-m missing argument"); }
        break;
      case 't':
        if ((--argc>0) && ((*++argv)[0]!='-')) {
          if (strcmp(*argv, "ivs") == 0) {
            ecs.ivsPush = True;
          } else if (strcmp(*argv, "psap") == 0) {
            ecs.ivsPush = False;
          }
          else { PrintUsage("-t invalid argument"); }
        }
        else { PrintUsage("-t missing argument"); }
        break;
      case 'c':
        if ((--argc>0) && ((*++argv)[0]!='-')) {
          ecs.campaign = *argv;
        } else { PrintUsage("-c missing argument"); }
        break;
      case 'i':
        if ((--argc>0) && ((*++argv)[0]!='-')) {
          ecs.inbase = *argv;
        } else { PrintUsage("-i missing argument"); }
        break;
      case 'o':
        if ((--argc>0) && ((*++argv)[0]!='-')) {
          ecs.outbase = *argv;
        } else { PrintUsage("-o missing argument"); }
        break;
      case 'p':
        if ((--argc>0) && ((*++argv)[0]!='-')) {
          ecs.portbase = *argv;
        } else { PrintUsage("-p missing argument"); }
        break;
      case 's':
        if ((--argc>0) && ((*++argv)[0]!='-')) {
          sscanf(*argv, "%d", &ecs.frameStart);
        } else { PrintUsage("-s missing argument"); }
        break;
      case 'x':
        if ((--argc>0) && ((*++argv)[0]!='-')) {
          sscanf(*argv, "%d", &ecs.frameStartHlack);
        } else { PrintUsage("-x missing argument"); }
        break;
      case 'h':
        PrintUsage(NULL);
        break;
      default:
        PrintUsage(NULL);
        break;
    }
  }
  if (ecs.mode     == Udef) { PrintUsage("-m option must be specified"); }
  if (ecs.campaign == NULL) { PrintUsage("-c option must be specified"); }
  if (ecs.inbase   == NULL) { PrintUsage("-i option must be specified"); }
}

void PrintUsage(const char *error) {
  printf("Usage:\n");
  printf("  standalone %s \\\n             %s\n\n",
    "-m <mode> [-t <trigger>] -c campaign -i inbase",
    "[-o outbase] [-p portbase] [-s startframe] [-x startframehlack]");

  printf("Choices for <mode>\n");
  printf("  * ivs    : IVS  normal operation      [ io = i pcm  o pcm  ]\n");
  printf("  * ivsrx  : IVS  receiver              [ io = i pcm  o port ]\n");
  printf("  * ivstx  : IVS  transmitter           [ io = i port o pcm  ]\n");
  printf("  * psap   : PSAP normal operation      [ io = i pcm  o pcm  ]\n");
  printf("  * psaprx : PSAP receiver              [ io = i pcm  o port ]\n");
  printf("  * psaptx : PSAP transmitter           [ io = i port o pcm  ]\n\n");

  printf("Choices for <trigger>\n");
  printf("  * ivs    : IVS  to trigger MSD transmission                 \n");
  printf("  * psap   : PSAP to trigger MSD transmission      [ default ]\n\n");

  printf("Other options\n");
  printf("  * -c campaign file                                          \n");
  printf("  * -i filename base input                                    \n");
  printf("  * -o filename base output                 [ default = NULL ]\n");
  printf("  * -p filename base portlog                [ default = NULL ]\n");
  printf("  * -s start of modem processing            [ default = 50   ]\n");
  printf("  * -x start of HLACK after MSD received    [ default = 10   ]\n\n");

  if (error != NULL) printf("Error: %s\n", error);
  exit(1);
}

void ConvertMsd(const char *msdChar, Ord8 *msd) {
  unsigned int n, x;

  for (n = 0; n < MSD_MAX_LENGTH; n++) {
    if (sscanf(&msdChar[n<<1], "%2x", &x) == 1) {
      msd[n] = x;
    } else {
      Abort("invalid MSD characters %c%c", msdChar[n<<1], msdChar[(n<<1)+1]);
    }
  }
}

/*============================================================================*/
/* File IO PCM data                                                           */
/*----------------------------------------------------------------------------*/

int GetBuffer(const char *base, const char *src, int index, int offset) {
  int  n;
  int  count = 0;
  char pcmname[501];
  FILE *pcmfile;

  /* File name */
  sprintf(pcmname, "%s%s%d.pcm", base, src, index);

  /* Open file */
  if ((pcmfile = fopen(pcmname, "rb")) == NULL) {
    Abort("unable to open file %s", pcmname);
  }
  else printf("opening %s\n", pcmname);

  /* Read PCM frames into buffer */
  for (;;) {
    if ((n = fread(&ecs.pcmBuffer[offset*PCM_LENGTH],
      sizeof(*ecs.pcmBuffer), PCM_LENGTH, pcmfile)) < PCM_LENGTH)
    {
      printf("reading %d frames (%d byte left)\n", count, n);
      break;
    } else {
      count++;
      offset++;
    }
    if (offset >= MAX_FRAMES) {
      Abort("PCM buffer overflow, increase MAX_FRAMES value!");
    }
  }
  /* Close file */
  if (pcmfile != NULL) fclose(pcmfile);
  return count;
}

int PutBuffer(const char *base, const char *dest, int index, int length) {
  int  n;
  char outname[501];
  FILE *outfile;

  /* File name */
  sprintf(outname, "%s%s%d.out", base, dest, index);

  /* Open file */
  if ((outfile = fopen(outname, "wb")) == NULL) {
    Abort("unable to open file %s", outname);
  }
  else printf("writing %s\n", outname);

  /* Write file */
  n = fwrite(ecs.pcmBuffer, sizeof(*ecs.pcmBuffer), length*PCM_LENGTH, outfile);

  /* Close file */
  if (outfile != NULL) fclose(outfile);
  return n;
}

/*============================================================================*/
/* File IO Port logging                                                       */
/*----------------------------------------------------------------------------*/
/* return:  >0 number of data items processed                                 */
/*          =0 end of file reached                                            */
/*          -1 error opening file                                             */
/*          -2 error in reading or writing data                               */
/*----------------------------------------------------------------------------*/

void PortLogInit(PortLog *log, const char *filename) {
  if (filename == NULL) {
    log->filename[0] = '\0';
  } else {
    strncpy(log->filename, filename, 500);
    log->filename[500] = '\0';
  }
  log->count = 0;
}

int PortLogWrite(PortLog *log, CtrlPort *port) {
  FILE *f;
  int n, rval;

  if (strlen(log->filename) == 0) {
    return 0;
  } else if (log->count > 0) {
    f = fopen(log->filename, "a");
  } else {
    f = fopen(log->filename, "w");
  }
  if (f == NULL) {
    return -1;
  } else {
    if (port->state == CtrlTxIdle || port->state == CtrlTxSend) {
      n = fprintf(f, "%05d  %-10s % d  %-14s % 3d  %-16s % 4d\n",
        ++log->count,
        "port.state", port->state,
        "port.tx.dlData", port->u.tx.dlData,
        "port.tx.dlIndex", port->u.tx.dlIndex);
    } else {
      n = fprintf(f, "%05d  %-10s % d  %-14s % 3d  %-16s % 4d\n",
        ++log->count,
        "port.state", port->state,
        "port.rx.dlData", port->u.rx.dlData,
        "port.rx.dlMetric", port->u.rx.dlMetric);
    }
    if (n < 0) {
      rval = -2;
    } else {
      rval = 4;
    }
    fclose(f);
    return rval;
  }
}

int PortLogRead(PortLog *log, CtrlPort *port) {
  FILE *f;
  int n, rval, state, value[2];

  if (strlen(log->filename) == 0) {
    return 0;
  } else if ((f = fopen(log->filename, "r")) == NULL) {
    return -1;
  } else {
    if (log->count > 0) {
      fsetpos(f, &log->filepos);
    }
    n = fscanf(f, "%d %*s %d %*s %d %*s %d\n",
      &log->count, &state, &value[0], &value[1]);

    if (state == CtrlTxIdle || state == CtrlTxSend) {
      port->state        = state;
      port->u.tx.dlData  = value[0];
      port->u.tx.dlIndex = value[1];
    } else {
      port->state         = state;
      port->u.rx.dlData   = value[0];
      port->u.rx.dlMetric = value[1];
    }
    if (n <= 0 && feof(f)) {
      rval = 0;
    } else if (n != 4) {
      rval = -2;
    } else {
      fgetpos(f, &log->filepos);
      rval = n;
    }
    fclose(f);
    return rval;
  }
}
