/*                                                    Version 3.02 - 15/Jun/2010
================================================================================

                           U    U   GGG    SSS  TTTTT
                           U    U  G      S       T
                           U    U  G  GG   SSS    T
                           U    U  G   G      S   T
                            UUUU    GGG    SSS    T

                    ========================================
                     ITU-T - USER'S GROUP ON SOFTWARE TOOLS
                    ========================================


COPYRIGHT NOTE:

This source code, and all of its derivations, is subject to the "ITU-T General
Public License". Please have it read in the distribution disk, or in the ITU-T
Recommendation G.191 on "SOFTWARE TOOLS FOR SPEECH AND AUDIO CODING STANDARDS".


MODULE: G711.C, G.711 ENCODING/DECODING FUNCTIONS

ORIGINAL BY:

Simao Ferraz de Campos Neto          Rudolf Hofmann
CPqD/Telebras                        PHILIPS KOMMUNIKATIONS INDUSTRIE AG
DDS/Pr.11                            Kommunikationssysteme
Rd. Mogi Mirim-Campinas Km.118       Thurn-und-Taxis-Strasse 14
13.085 - Campinas - SP (Brazil)      D-8500 Nuernberg 10 (Germany)

Phone : +55-192-39-6396              Phone : +49 911 526-2603
FAX   : +55-192-53-4754              FAX   : +49 911 526-3385
EMail : tdsimao@venus.cpqd.ansp.br   EMail : HF@PKINBG.UUCP


FUNCTIONS:

alaw_compress: ... compands 1 vector of linear PCM samples to A-law;
                   uses 13 Most Sig. Bits (MSBs) from input and
                   8 Least Sig. Bits (LSBs) on output.

alaw_expand: ..... expands 1 vector of A-law samples to linear PCM;
                   uses 8 Least Sig. Bits (LSBs) from input and
                   13 Most Sig. Bits (MSBs) on output.

ulaw_compress: ... compands 1 vector of linear PCM samples to u-law;
                   uses 14 Most Sig. Bits (MSBs) from input and
                   8 Least Sig. Bits (LSBs) on output.

ulaw_expand: ..... expands 1 vector of u-law samples to linear PCM;
                   uses 8 Least Sig. Bits (LSBs) from input and
                   14 Most Sig. Bits (MSBs) on output.

PROTOTYPES: ...... in g711.h


HISTORY:
--/Apr/1991  1.0   First version of the G711 module
10/Dec/1991  2.0   Break-up in individual functions for A/u-law;
                   correction of bug in compression routines
                   (use of 1 and 2 complement); demo program inside module.
08/Feb/1992  3.0   Demo as separate file;
31/Jan/2000  3.01  Updated documentation text; no change in functions
                   <simao.campos@labs.comsat.com>
15/Jun/2010  3.02  Minor code reformatting
                   (for use with 3GPP TS 26.268 eCall reference source code)

================================================================================
*/

#include "g711.h"

/*
================================================================================

FUNCTION NAME:  alaw_compress
DESCRIPTION:    A-law encoding rule according ITU-T Rec. G.711.
PROTOTYPE:      void alaw_compress(long lseg, short *linbuf, short *logbuf)

PARAMETERS:
  lseg:   (In)  number of samples
  linbuf: (In)  buffer with linear samples
                (only 12 MSBits are taken into account)
  logbuf: (Out) buffer with compressed samples
                (8 bit right justified, without sign extension)

RETURN VALUE:   none.

HISTORY:
10/Dec/91  1.0  Separated A-law compression function

================================================================================
*/
void alaw_compress(long lseg, short *linbuf, short *logbuf)
{
  short ix, iexp;
  long  n;

  for (n = 0; n < lseg; n++) {
    /* 0 <= ix < 2048 (1's complement for negative values) */
    ix = (linbuf[n] < 0) ? (~linbuf[n]) >> 4 : (linbuf[n]) >> 4;

    /* do more if exponent > 0 */
    if (ix > 15) {                /* exponent = 0 for ix <= 15 */
      iexp = 1;                   /* first step: */
      while (ix > 16 + 15) {      /* find mantissa and exponent */
        ix >>= 1;
        iexp++;
      }
      ix -= 16;                   /* second step: remove leading '1' */
      ix += iexp << 4;            /* now compute encoded value */
    }
    if (linbuf[n] >= 0) {
      ix |= (0x0080);             /* add sign bit */
    }
    logbuf[n] = ix ^ (0x0055);    /* toggle even bits */
  }
}

/*
================================================================================

FUNCTION NAME:  alaw_expand
DESCRIPTION:    A-law decoding rule according ITU-T Rec. G.711.
PROTOTYPE:      void alaw_expand(long lseg, short *logbuf, short *linbuf)

PARAMETERS:
  lseg:   (In)  number of samples
  logbuf: (In)  buffer with compressed samples
                (8 bit right justified, without sign extension)
  linbuf: (Out) buffer with linear samples
                (13 bits left justified)

RETURN VALUE:   none.

HISTORY:
10/Dec/91  1.0  Separated A-law expansion function

================================================================================
*/
void alaw_expand(long lseg, short *logbuf, short *linbuf)
{
  short ix, mant, iexp;
  long  n;

  for (n = 0; n < lseg; n++) {
    ix = logbuf[n] ^ (0x0055);    /* re-toggle toggled bits */

    ix &= (0x007F);               /* remove sign bit */
    iexp = ix >> 4;               /* extract exponent */
    mant = ix & (0x000F);         /* now get mantissa */

    if (iexp > 0) {
      mant = mant + 16;           /* add leading '1' if exponent > 0 */
    }
    /* now mantissa left justified and 1/2 quantization step added */
    mant = (mant << 4) + (0x0008);

    if (iexp > 1) {
      mant = mant << (iexp - 1);  /* now left shift according exponent */
    }
    /* invert if negative sample */
    linbuf[n] = (logbuf[n] > 127) ? mant : -mant;
  }
}

/*
================================================================================

FUNCTION NAME:  ulaw_compress
DESCRIPTION:    u-law encoding rule according ITU-T Rec. G.711.
PROTOTYPE:      void ulaw_compress(long lseg, short *linbuf, short *logbuf)

PARAMETERS:
  lseg:   (In)  number of samples
  linbuf: (In)  buffer with linear samples
                (only 12 MSBits are taken into account)
  logbuf: (Out) buffer with compressed samples
                (8 bit right justified, without sign extension)

RETURN VALUE:   none.

HISTORY:
10/Dec/91  1.0  Separated u-law compression function

================================================================================
*/
void ulaw_compress(long lseg, short *linbuf, short *logbuf)
{
  short i;            /* aux.var. */
  short absno;        /* absolute value of linear (input) sample */
  short segno;        /* segment (Table 2/G711, column 1) */
  short low_nibble;   /* low  nibble of log companded sample */
  short high_nibble;  /* high nibble of log companded sample */
  long  n;

  for (n = 0; n < lseg; n++) {
    /* change from 14 bit left justified to 14 bit right justified */
    /* compute absolute value; adjust for easy processing */

    /* compute 1's complement in case of negative samples */
    /* 33 is the difference value between the thresholds for A-law and u-law */
    absno = (linbuf[n] < 0)
      ? 33 + ((~linbuf[n]) >> 2)
      : 33 + ((linbuf[n]) >> 2);

    /* limitation to "absno" < 8192 */
    if (absno > (0x1FFF)) {
      absno = (0x1FFF);
    }
    /* determination of sample's segment */
    i = absno >> 6;
    segno = 1;
    while (i != 0) {
      segno++;
      i >>= 1;
    }
    /* mounting the high-nibble of the log-PCM sample */
    high_nibble = (0x0008) - segno;

    /* mounting the low-nibble of the log PCM sample */
    /* right shift of mantissa and masking away leading '1' */
    low_nibble = (absno >> segno) & (0x000F);
    low_nibble = (0x000F) - low_nibble;

    /* joining the high-nibble and the low-nibble of the log PCM sample */
    logbuf[n] = (high_nibble << 4) | low_nibble;

    /* add sign bit */
    if (linbuf[n] >= 0) {
      logbuf[n] = logbuf[n] | (0x0080);
    }
  }
}

/*
================================================================================

FUNCTION NAME:  ulaw_expand
DESCRIPTION:    u-law decoding rule according ITU-T Rec. G.711.
PROTOTYPE:      void ulaw_expand(long lseg, short *logbuf, short *linbuf)

PARAMETERS:
  lseg:   (In)  number of samples
  logbuf: (In)  buffer with compressed samples
                (8 bit right justified, without sign extension)
  linbuf: (Out) buffer with linear samples
                (14 bits left justified)

RETURN VALUE:   none.

HISTORY:
10/Dec/91  1.0  Separated u-law expansion function

================================================================================
*/
void ulaw_expand(long lseg, short *logbuf, short *linbuf)
{
  short segment;      /* segment (Table 2/G711, column 1) */
  short mantissa;     /* low  nibble of log companded sample */
  short exponent;     /* high nibble of log companded sample */
  short sign;         /* sign of output sample */
  short step;
  long  n;

  for (n = 0; n < lseg; n++) {
    /* sign-bit = 1 for positive values */
    sign = (logbuf[n] < (0x0080)) ? -1 : 1;

    mantissa = ~logbuf[n];                  /* 1's complement of input value */
    exponent = (mantissa >> 4) & (0x0007);  /* extract exponent */
    segment  = exponent + 1;                /* compute segment number */
    mantissa = mantissa & (0x000F);         /* extract mantissa */

    /* compute quantized sample (14 bit left justified!) */
    /* position of the LSB = 1 quantization step */
    step = (4) << segment;

    linbuf[n] = sign *                      /* sign */
      (((0x0080) << exponent)               /* '1' preceding the mantissa */
      + step * mantissa                     /* left shift of mantissa */
      + step / 2                            /* 1/2 quantization step */
      - 4 * 33);
  }
}
